C          *************************************************************
C          *                                                           *
C          *     VMP$ - Vicki Woolf file mapping services library      *
C          *                                                           *
C          *************************************************************



C***********************************************************************
C
C	VICKI W. WOOLF
C	PROGRAM OF COMPUTER GRAPHICS, 120 RAND HALL, CORNELL UNIVERSITY
C	ITHACA, NEW YORK  14853   (607) 256-7444
C
C		This library consists of set of fortran-callable
C		subroutines that enable disk files to be treated
C		as program sections. The main purpose is to bypass 
C		RMS I/O.  This should decrease the I/O time of the
C		program substantially.
C
C		There are three main entry points:
C			VMP$MAP_FILE
C			VMP$MAP_RECORD
C			VMP$UNMAP_FILE
C
C		MAP_FILE accesses the file according to the name,
C		type, and access method specified by the caller.
C		The routine expands the user's 	virtual address
C		space by the number of bytes specified as window
C		size plus 1 block, and returns to the user an internal
C		file identifier for reference purposes in future calls
C		to MAP_RECORD and UNMAP_FILE.  Because virtual
C		address space is allocated as each file is referenced
C		the total number of files is limited only by the
C		user quotas PGFLQUOTA and WSQUOTA.  MAP_FILE increases
C		the user's virtual address space, but does not map
C		the file.  New files are opened with the user attributes
C		of sequential files with variable length records.
C		No maximum record size is specified.  This should not
C		create a problem.
C		MAP_RECORD uses the input parameters RECPOSITION
C		and RECLENGTH to determine whether or not that record
C		is currently in the virtual address space (or window)
C		for the requested file.  Note: RECLENGTH must be
C		less that the window size or if RECPOSITION is not
C		on a page boundary then RECLENGTH must be less than
C		WINDOWSIZE - (offset RECPOSITION is into a page).
C		If the record is not currently in the window than
C		the window is remapped to overlay the requested portion
C		of the file.  If necessary the file is extended to
C		handle the requested record.  Once the window overlays
C		the file appropriately the subroutine returns to the
C		caller passing the caller a pointer to where the
C		data is in virtual space.
C		UNMAP_FILE is used to deaccess the file, truncating it
C		if necessary and freeing up the virtual address space.
C		For a read/write file, the calling program can specify
C		the filelength to be used.  
C	
C
C***********************************************************************

C*********** VMP$MAP_FILE *******************************************
C
C	VMP$MAP_FILE opens the file specified by FILESPEC and returns
C	to the caller a unique identification number and the total
C	number of bytes in the file.  (filelength=0, if the file is new.)
C	VMP$MAP_FILE increases the users's virtual address space by
C	enough blocks to cover the windowsize requested plus 1.



	INTEGER*4	FUNCTION VMP$MAP_FILE(ID,FILESPEC,TYPE,ACCESS
	1			,WINDOWSIZE,FILELENGTH)

	INCLUDE		'VMPLIBDEF.FOR/NOLIST'

	EXTERNAL	SS$_NORMAL
	INTEGER*4	SS$_NORMAL
	INTEGER*4	SYS$EXPREG
	INTEGER*4	VMP$CR8_FILDSCR,VMP$OPEN_NEW,VMP$OPEN_OLD
	INTEGER*4	VMP$INIT_DESCR
	INTEGER*4	VMP$COPY,VALUE_OF


	INTEGER*4	ID,WINDOWSIZE,FILELENGTH
	CHARACTER*(*)	FILESPEC,TYPE,ACCESS
	INTEGER*4	ISTATUS			! local


C********** Explanation of parameters and locals **********
C
C	ID		(parameter - longword - integer) Internal
C			file identifier returned to caller.  This
C			number should be used by the caller for
C			all calls to MAP_RECORD, and UNMAP_FILE
C			that relate to this file.
C	WINDOWSIZE	(parameter - longword - integer) Number of
C			bytes for mapping window.  Window will actually
C			be the number of blocks needed to handle the
C			size requested.
C	FILELENGTH	(parameter - longword - integer) Number of
C			bytes in the file. This is calculated from
C			the user attributes area of the file header.
C			(end-of-file marker *512) + first-available-
C			byte number.)
C	FILESPEC	(parameter - character string) Name of file
C			provided by caller. This can be a logical
C			name.  If no directory is provided the default
C			directory is assumed.
C	TYPE		(parameter - character string) File type.
C			'NEW'(create) or 'OLD'(open).
C	ACCESS		(parameter - character string) File access mode.
C			'R/O' (readonly - for old files only) or 
C			'R/W' (read/write).
C	ISTATUS		(local - longword - integer). Program status variable.





C**********  Open file - create file descriptor block **********
C
C  >Initialize temporary file descriptor block.
C	>If not successful-Return error status.
C  >Open (old) or create (new) file.
C	>If not successful-Return error status.
C  >Expand virtual address space by a windowsize to accomodate file.
C 	>If not successful-Return error status.
C  >Set MAP$START_BYTE to -1 to indicate that file has not been mapped.
C  >Create a file descriptor block for file.
C	>If not successful-Return error status.
C  >Copy temporary file descriptor block to permanent one.
C  >Return fileid, filelength, and successful status to caller.

	ISTATUS = VMP$INIT_DESCR(TYPE,ACCESS,WINDOWSIZE)
	IF (.NOT. ISTATUS) THEN
		CALL REPORT('VMP$MAP_FILE ... invalid parameter '''
	1		//NCODE(ISTATUS)//'''X')		
		VMP$MAP_FILE = ISTATUS
		RETURN
	ENDIF



	IF (MAP$TYPE .EQ. MAP$C_NEW) THEN
		ISTATUS = VMP$OPEN_NEW(FILESPEC(1:LENGTH(FILESPEC))
	ELSE
		ISTATUS = VMP$OPEN_OLD(FILESPEC(1:LENGTH(FILESPEC))
	ENDIF

	IF (ISTATUS .EQ. VMP$_INVFIL) THEN
		VMP$MAP_FILE = VMP$_OPEFAI
	ELSEIF (.NOT. ISTATUS) THEN
		CALL REPORT('VMP$MAP_FILE .. open failure ')
		VMP$MAP_FILE = ISTATUS
	ENDIF


	ISTATUS =  SYS$EXPREG(%VAL(MAP$WINDOW_BLKS)
	1		,MAP$MAP_ADDR
	1		,%VAL(3)
	1		,%VAL(0)
	1		)
	IF (.NOT. ISTATUS) THEN
		CALL REPORT(
	1	 'VMP$MAP_FILE .. failure to expand program region '''
	2	 //NCODE(ISTATUS)//'''X')
		VMP$MAP_FILE = VMP$_VAFAI
		RETURN
	ENDIF
	MAP$START_BYTE = -1

	ISTATUS = VMP$CR8_FILDSCR(MAP$FILEID)
	IF (.NOT. ISTATUS) THEN
		CALL REPORT(
	1	 'VMP$MAP_FILE .. failure to expand program region '''
	2	 //NCODE(ISTATUS)//'''X')
		VMP$MAP_FILE = VMP$_VAFAI
		RETURN
	ENDIF

	DESCR(1) = FILE_DESCR_LEN
	DESCR(2) = VALUE_OF(%VAL(CURR_PTR))
	CALL VMP$COPY(FILE_DESCR_C,DESCR)

	FILELENGTH = MAP$FILE_LEN
	ID = MAP$FILEID
	VMP$MAP_FILE = %LOC(SS$_NORMAL)
	RETURN
	END

C********** VMP$MAP_RECORD **********
C
C	For any specified recposition, this procedure returns a
C	virtual address through which at least reclength number of
C	bytes starting at that position may be accessed by the 
C	caller.  For file reading, this length is set by the 
C	caller to the largest record expected to occur in the file.
C	For writing, it is set to the actual length of the record
C	being written.  If the specified position is beyond the
C	current end-of-file, the file is extended as necessary
C	(subject to access flag.)

	INTEGER*4	FUNCTION VMP$MAP_RECORD(ID,RECPOSITION,RECLENGTH
	1			,ADDRESS)


	INCLUDE		'VMPLIBDEF.FOR/NOLIST'

	PARAMETER	PAGE_MASK = '000001FF'X

	INTEGER*4	ID,RECPOSITION,RECLENGTH,ADDRESS
	INTEGER*4	ISTATUS			! local

	EXTERNAL	SS$_NORMAL,SS$_ENDOFFILE,SS$_CREATED
	INTEGER*4	SS$_NORMAL,SS$_ENDOFFILE,SS$_CREATED

	INTEGER*4	SYS$CRMPSC,SYS$EXPREG,SYS$DELTVA
	INTEGER*4	VMP$EXTEND,VMP$GET_FILDSCR,VMP$UPDATE
	INTEGER*4	VALUE_OF



C********** Explanation of locals and parameters ***********
C
C	ID		= internal file identifier number that
C			  was returned to the caller from MAP_FILE.
C	RECPOSITION	= byte offset from start of file that the
C			  caller requests.
C	RECLENGTH	= number of bytes. RECPOSITION + RECLENGTH
C			  must be able to fit within the window.
C	ADDRESS		= pointer to virtual address space where
C			  data specified by RECPOSITION and RECLENGTH
C			  exists.



C********** verify that record requested is in window **********
C
C  >If reclength is <= 0 or too large for window:
C	>Return error status.
C  >If no file-descriptor exists for requested file id:
C	>Return error status.
C  >If record requested is not in current window:
C	>If record is beyond end-of-file:
C		>Extend file.
C		>If extend fails - (readonly file,...):
C			>Return error status.
C	>Update section file on disk.
C	>If update fails:
C		>Return error status.
C	>Logically overlay window on file.
C	>Remap section file.
C	>If section mapping fails:
C		>Return error status.
C  >Calculate address of requested record
C  >Move file descriptor block from temporary storage
C  >Return true status.

	IF (RECLENGTH .LE. 0 
	1 .OR. (RECLENGTH 
	2   .GT. (MAP$WINDOW_SIZE - (RECPOSITION .AND. PAGE_MASK)))
	3 .OR. RECPOSITION .LT. 0) THEN
		VMP$MAP_RECORD = VMP$_INVPAR
		RETURN
	ENDIF

	ISTATUS = VMP$GET_FILDSCR(ID)
	IF (.NOT. ISTATUS) THEN
		VMP$MAP_RECORD = ISTATUS
		RETURN
	ENDIF	

	IF (MAP$START_BYTE .LT. 0
	1  .OR. .NOT. ( (RECPOSITION .GE. MAP$START_BYTE)
	1	.AND. (RECPOSITION + RECLENGTH-1  .LT.
	1	MIN(MAP$EXT_FILELEN,MAP$START_BYTE + MAP$WINDOW_SIZE - 1))) ) 
	1	THEN

		IF (RECPOSITION + RECLENGTH - 1 .GT. MAP$EXT_FILELEN) THEN
			IF (MAP$ACCESS_TYPE .EQ. MAP$C_RW) THEN
				ISTATUS = VMP$EXTEND()
				IF (.NOT. ISTATUS) THEN
					CALL REPORT(
	1				  'VMP$MAP_RECORD .. failure'
	2				  //' to extend file '''
	3				  // NCODE(ISTATUS)//'''X')
					VMP$MAP_RECORD = VMP$_EXTFAI
					RETURN
				ENDIF
			ELSE
				CALL REPORT('VMP$MAP_RECORD .. attempt'
	1			  //' to extend a readonly file')
				VMP$MAP_RECORD = VMP$_NOWRITE
				RETURN
			ENDIF
		ENDIF

		IF (MAP$ACCESS_TYPE .EQ. MAP$C_RW) THEN
			ISTATUS = VMP$UPDATE()
			IF (.NOT. ISTATUS) THEN
				CALL REPORT(
	1			  'VMP$MAP_RECORD .. failure'
	2			  //' to update section file '''
	3			  // NCODE(ISTATUS)//'''X')
				VMP$MAP_RECORD = VMP$_UPDFAI
				RETURN
			ENDIF
		ENDIF


		MAP$START_BYTE = RECPOSITION .AND. .NOT.PAGE_MASK
		ITEMP = 0
		IF (MAP$ACCESS_TYPE .EQ. MAP$C_RW) ITEMP = SEC$M_WRT


		ISTATUS = SYS$CRMPSC(MAP$MAP_ADDR,RETADR,%VAL(3)
	1		,%VAL(ITEMP),,,
	1		,%VAL(MAP$CHANNEL)
	1		,%VAL(MAP$WINDOW_BLKS)
	1		,%VAL(MAP$START_BYTE/512 + 1),,)
		IF ( .NOT.(
	1	(ISTATUS .EQ. %LOC(SS$_ENDOFFILE).AND.MAP$NO_EXTENDS.GE.1)
	1	.OR. ISTATUS .EQ. %LOC(SS$_NORMAL) 
	1	.OR. ISTATUS .EQ. %LOC(SS$_CREATED) ) ) THEN
			CALL REPORT('VMP$MAP_RECORD .. failure to'
	1		  //' map section file '''//NCODE(ISTATUS)//'''X')
			VMP$MAP_RECORD = VMP$_MAPFAI
			RETURN
		ENDIF

		DESCR(1) = FILE_DESCR_LEN
		DESCR(2) = VALUE_OF(%VAL(CURR_PTR))
		CALL VMP$COPY(FILE_DESCR_C,DESCR)

	ENDIF

	ADDRESS = MAP$MAP_ADDR(1) + RECPOSITION - MAP$START_BYTE
	VMP$MAP_RECORD = %LOC(SS$_NORMAL)

	RETURN
	END

C*********** VMP$UNMAP_FILE ***************************************
C
C	Close file, truncating it if necessary.  Delete virtual 
C	address space of file and rebuild linked list of file
C	descriptor blocks.
	
	INTEGER*4	FUNCTION VMP$UNMAP_FILE(ID,FILELENGTH)

	INCLUDE		'VMPLIBDEF.FOR/NOLIST'

	INTEGER*4	ID,FILELENGTH
	INTEGER*4	ISTATUS				! local

	EXTERNAL	SS$_NORMAL
	INTEGER*4	SS$_NORMAL,SYS$DELTVA
	INTEGER*4	VMP$GET_FILDSCR,VMP$TRUNCATE,VMP$COPY,VMP$UPDATE
	INTEGER*4	VMP$CLOSE
	INTEGER*4	STORE_VALUE,VALUE_OF



C*********** Explanation of parameters and locals ***********
C
C	ID		= internal file identifier returned to caller
C			  from call to MAP_FILE.
C	FILELENGTH	= number of bytes in file.  This parameter
C			  should be used for read/write files only.
C	ISTATUS		= (local) - program status longword.



C*********** undate file, close and truncate (if necessary),delete window ****
C
C  >Get file descriptor block - return error if fails.
C  >Update section file on disk - return error if fails.
C  >Delete section by deleting pages associated with it.
C  >Close file - return error if fails.
C  >Rebuild linked list of file descriptor blocks.
C  >Delete address space of descriptor block - return error if fails.
C  >Return true status.

	ISTATUS = VMP$GET_FILDSCR(ID)
	IF (.NOT. ISTATUS) THEN
		CALL REPORT('VMP$UNMAP_FILE .. file has not been mapped')
		VMP$UNMAP_FILE = ISTATUS
		RETURN
	ENDIF

	ISTATUS = SYS$DELTVA(MAP$MAP_ADDR,,%VAL(3))
	IF (.NOT. ISTATUS) THEN
		CALL REPORT('VMP$UNMAP_FILE .. '
	1		//'delete virtual address space failure '''
	2		//NCODE(ISTATUS)//'''X')
		VMP$UNMAP_FILE = VMP$_DVAFIL
		RETURN
	ENDIF

	ISTATUS = VMP$CLOSE(FILELENGTH)
	IF (.NOT. ISTATUS) THEN
		CALL REPORT('VMP$UNMAP_FILE .. close failure')
		VMP$UNMAP_FILE = ISTATUS
		RETURN
	ENDIF

	INADR(1) = VALUE_OF(%VAL(CURR_PTR))
	INADR(2) = VALUE_OF(%VAL(CURR_PTR))
	IF (PREV_PTR .NE. 0) THEN
		CALL STORE_VALUE(%VAL(PREV_PTR),MAP$NEXT_FILE)
	ELSE
		CALL STORE_VALUE(MAP_PTR,MAP$NEXT_FILE)
	ENDIF
	PREV_PTR = 0
	CURR_PTR = %LOC(MAP_PTR)

	MAP$NEXT_FILE = 0
	MAP$FILEID = 0

	ISTATUS = SYS$DELTVA(INADR,,%VAL(3))
	IF (.NOT. ISTATUS) THEN
		CALL REPORT('VMP$UNMAP_FILE .. delete virtual address space'''
	1		//NCODE(ISTATUS)//'''X')
		VMP$UNMAP_FILE = ISTATUS
		RETURN
	ENDIF

	VMP$UNMAP_FILE = %LOC(SS$_NORMAL)
	RETURN
	END

C********** VMP$OPEN_NEW ******************************************

	INTEGER*4	FUNCTION VMP$OPEN_NEW(FILESPEC)

	INCLUDE		'VMPLIBDEF.FOR/NOLIST'

	CHARACTER*(*)	FILESPEC

	INTEGER*4	ISTATUS			! local

	INTEGER*4	VW$GET_FID
	EXTERNAL	SS$_NORMAL,IO$_CREATE,IO$M_CREATE,IO$M_ACCESS
	EXTERNAL	IO$M_DLOCK,IO$_MODIFY
	INTEGER*4	SS$_NORMAL,IO$_CREATE,IO$M_CREATE,IO$M_ACCESS
	INTEGER*4	IO$M_DLOCK,IO$_MODIFY
	INTEGER*4	SYS$ASSIGN,SYS$QIOW



C********** Explanation of parameters and locals **********
C
C	FILESPSEC	= character string name of file to be 
C			  created.
C	ISTATUS		= local - program status longword



C********** Create a new file *********
C
C  >Set up FIB 
C  >Get full filename.
C  >If not successful:
C	>Return error status.
C  >Parse directory name from resultant filename string.
C  >Create a .DIR filename.
C  >Get FID (file identification number) for the directory file.
C  >Parse device name from resultant name string.
C  >Assign a channel to the device.
C  >If not successful,then:
C	>Return error status code.
C  >Set up FIB (File Information Block).
C  	extend-size = number of window blocks
C	request new version
C	if access is read/write request write access
C  >Create file (QIOW).
C  >If not successful:
C	>Return error status.
C  >Update file descriptor block and record attribute information.
C  >Update file header (QIOW).
C  >If not successful:
C	>Return error status.
C  >Return success.

	FIB$W_DID_NUM = 0
	FIB$W_DID_SEQ = 0
	FIB$W_DID_RVN = 0
	FIB$W_FID_NUM = 0
	FIB$W_FID_SEQ = 0
	FIB$W_FID_RVN = 0

	ISTATUS = VW$GET_FID(MAP$FAB		! BUFFER FOR FAB (TEMP)
	1	,FILESPEC
	1	,MAP$FILESPEC
	1	,MAP$FILESPECLEN
	1	,FIB$W_DID
	1	,FIB$W_FID
	1	)

	IPOS_END = INDEX(MAP$FILESPEC,']')
	IPOS = IPOS_END
   10	IPOS = IPOS - 1

	IF (IPOS .LE. 0)
	1   THEN
		VMP$OPEN_FILE = VMP$_OPEFAI
		RETURN
	ENDIF

	IF (MAP$FILESPEC(IPOS:IPOS) .EQ. '.')
	1   THEN
		CALL VW$GET_FID(MAP$FAB
	1		,MAP$FILESPEC(1:IPOS-1)//']'			
	1			//MAP$FILESPEC(IPOS+1:IPOS_END-1)
	1			//'.DIR'
	1		,TEMP_BUFFER_C
	1		,ITEMP
	1		,TEMP_BUFFER(6)
	1		,FIB$W_DID
	1		)
	    ELSE IF (MAP$FILESPEC(IPOS:IPOS).EQ. '[')
	1     THEN
		CALL VW$GET_FID(MAP$FAB
	1			,MAP$FILESPEC(1:IPOS-1)//'[000000]'
	1			//MAP$FILESPEC(IPOS+1:IPOS_END-1)
	1			//'.DIR'
	1		,TEMP_BUFFER_C
	1		,ITEMP
	1		,TEMP_BUFFER(6)
	1		,FIB$W_DID
	1		)
	    ELSE
		GO TO 10
	ENDIF

	IPOS = INDEX(MAP$FILESPEC,':')
	ISTATUS = SYS$ASSIGN(MAP$FILESPEC(1:IPOS),MAP$CHANNEL
	1		,%VAL(3),)
	IF (.NOT. ISTATUS) THEN
		CALL REPORT('VMP$OPEN_NEW .. failure to assign channel '''
	1	  //NCODE(ISTATUS)//'''X')
		VMP$OPEN_NEW=VMP$_OPNFIL
		RETURN
	ENDIF


	FIB$L_ACCTL = FIB$M_DLOCK .OR. FIB$M_WRITE
	FIB$W_EXCTL =  FIB$M_EXTEND .OR. FIB$M_ALCONB
	FIB$L_EXSZ = 0
	FIB$L_EXVBN = 0
	FIB$W_NMCTL = FIB$M_NEWVER
	FIB$W_FID_NUM = 0
	FIB$W_FID_SEQ = 0
	FIB$W_FID_RVN = 0

	ISTATUS = SYS$QIOW( 
	1		,%VAL(MAP$CHANNEL)
	1		,%VAL(%LOC(IO$_CREATE) .OR. %LOC(IO$M_CREATE)
	1		   .OR. %LOC(IO$M_ACCESS)  )
	1		,IOSB
	1		,
	1		,
	1		,MAP$FIB_DESCR
	1		,%DESCR(MAP$FILESPEC(INDEX(MAP$FILESPEC,']')+1:
	1			INDEX(MAP$FILESPEC,';')-1) )
	1		,,
	1		,
	1		,)

	IF (.NOT. ISTATUS .OR. .NOT. IOSB_W(1))
	1   THEN
		CALL REPORT('VMP$OPEN_NEW .. failure to create file '''
	1		//NCODE(IOSB_W(1))//'''X')
		VMP$OPEN_NEW =  VMP$_OPNFIL
		RETURN
	ENDIF

	MAP$ALLOC_BLKS = FIB$L_EXSZ
	MAP$EOF_BLK_NO = 0
	MAP$FREE_BYTE  = 0
	MAP$FILE_LEN   = 0
	MAP$EXT_FILELEN= 0

	MAP$W_REC_ALLOC = MAP$ALLOC_BLKS
	MAP$W_REC_EOF	= MAP$EOF_BLK_NO
	MAP$W_REC_FREE  = MAP$FREE_BYTE
	MAP$B_REC_TYPE  = FHD$VARLEN		! Variable length records.


	FIB$W_EXCTL = 0
	FIB$L_EXVBN = 0
	FIB$L_EXSZ  = 0
	FIB$W_NMCTL =FIB$M_FINDFID

	ISTATUS = SYS$QIOW( 
	1		,%VAL(MAP$CHANNEL)
	1		,IO$_MODIFY
	1		,IOSB
	1		,
	1		,
	1		,MAP$FIB_DESCR
	1		,
	1		,,
	1		,ATR$L_LIST
	1		,)

	IF (.NOT. ISTATUS .OR. .NOT. IOSB_W(1)) THEN
		CALL REPORT('VMP$OPEN_NEW .. failure to update file header '''
	1	  //NCODE(IOSB_W(1))//'''X')
		VMP$OPEN_NEW = VMP$_OPNFIL
	ELSE
		VMP$OPEN_NEW = %LOC(SS$_NORMAL)
	ENDIF
	RETURN
	END

C********** VMP$OPEN_OLD - access an old file ***************************

	INTEGER*4	FUNCTION VMP$OPEN_OLD(FILESPEC)

	INCLUDE		'VMPLIBDEF.FOR/NOLIST'

	CHARACTER*(*)	FILESPEC
	INTEGER*4	ISTATUS			! locals

	EXTERNAL	SS$_NORMAL,IO$_ACCESS,IO$M_ACCESS,IO$_MODIFY
	INTEGER*4	SS$_NORMAL,IO$_ACCESS,IO$M_ACCESS,IO$_MODIFY
	INTEGER*4	SYS$ASSIGN,SYS$QIOW
	INTEGER*4	VW$GET_FID



C********** Explanation of parameters and locals **********
C
C	FILESPEC	= character string variable that contains
C			  name of file to be opened.
C	ISTATUS		= local - program status.
C


C********** Accessing a old file **********
C
C  >Set up FIB
C  >Retrieve full file specification string,file id, and directory id.
C  >Assign a channel to the device name.
C  >If not successful.
C	>Return error status.
C  >Set access type.
C  >Create file (QIOW).
C  >If not successful:
C	>Return error status.
C  >Update fields for file descriptor update.
C  >Write record attributes into file header (QIOW.)
C  >If not successful:
C	>Return error status.
C  >If successful:
C	>Return true status.

	FIB$W_DID_NUM = 0
	FIB$W_DID_SEQ = 0
	FIB$W_DID_RVN = 0
	FIB$W_FID_NUM = 0
	FIB$W_FID_SEQ = 0
	FIB$W_FID_RVN = 0

	ISTATUS = VW$GET_FID(MAP$FAB		! BUFFER FOR FAB (TEMP)
	1	,FILESPEC
	1	,MAP$FILESPEC
	1	,MAP$FILESPECLEN
	1	,FIB$W_DID
	1	,FIB$W_FID
	1	)

	IF (.NOT. ISTATUS) THEN
		VMP$OPEN_OLD = VMP$_OPEFAI
		RETURN
	ENDIF

	IPOS = INDEX(MAP$FILESPEC,':')
	ISTATUS = SYS$ASSIGN(MAP$FILESPEC(1:IPOS),MAP$CHANNEL
	1		,%VAL(3),)
	IF (ISTATUS .NE. %LOC(SS$_NORMAL)) THEN
		CALL REPORT('VMP$OPEN_OLD .. failure to assign channel '''
	1	 //NCODE(ISTATUS)//'''X')
		VMP$OPEN_OLD = VMP$_OPNFIL
		RETURN
	ENDIF

	FIB$L_ACCTL = 0
	IF (MAP$ACCESS_TYPE .EQ. MAP$C_RW) FIB$L_ACCTL = 
	1		FIB$M_DLOCK .OR. FIB$M_WRITE
	FIB$W_EXCTL = 0			
	FIB$L_EXSZ = MAP$WINDOW_BLKS
	FIB$L_EXVBN = 0


	ISTATUS = SYS$QIOW( 
	1		,%VAL(MAP$CHANNEL)
	1		,%VAL( %LOC(IO$_ACCESS) .OR. %LOC(IO$M_ACCESS) )
	1		,IOSB
	1		,
	1		,
	1		,MAP$FIB_DESCR
	1		,%DESCR(MAP$FILESPEC(INDEX(MAP$FILESPEC,']')+1 :
	1				INDEX(MAP$FILESPEC,';')-1))
	1		,,
	1		,ATR$L_LIST
	1		,)


	MAP$ALLOC_BLKS	= MAP$W_REC_ALLOC
	MAP$EOF_BLK_NO  = MAP$W_REC_EOF
	MAP$FREE_BYTE   = MAP$W_REC_FREE
	MAP$FILE_LEN    = (MAP$EOF_BLK_NO - 1) * 512 + MAP$FREE_BYTE
	MAP$EXT_FILELEN = MAP$FILE_LEN


	IF (.NOT. ISTATUS .OR. .NOT. IOSB_W(1)) THEN
		CALL REPORT('VMP$OPEN_OLD .. failure to open file '''
	1	  //NCODE(IOSB_W(1))//'''X')
		VMP$OPEN_OLD = VMP$_OPNFIL
	ELSE
		VMP$OPEN_OLD = %LOC(SS$_NORMAL)
	ENDIF

	RETURN
	END

C********* VMP$CLOSE - deaccess file ***************************

	INTEGER*4	FUNCTION VMP$CLOSE(FILELENGTH)

	INCLUDE		'VMPLIBDEF.FOR/NOLIST'

	INTEGER*4	FILELENGTH		! parameter
	INTEGER*4	ISTATUS			! local

	EXTERNAL	SS$_NORMAL,IO$_DEACCESS,IO$_MODIFY
	INTEGER*4	SS$_NORMAL,IO$_DEACCESS,IO$_MODIFY
	INTEGER*4	SYS$QIOW,SYS$DASSGN
	INTEGER*4	VMP$TRUNCATE



C********** Explanation of Parameters and Locals **********
C
C	FILELENGTH 	Requested length of file in bytes.
C			This parameter overrides the current
C			length of file for all read/write files.
C
C	ISTATUS		(local). Program module error status.



C********** Close file ***********
C
C  >If file access type is read/write then:
C 	>If filelength is not equal to original filelength then:
C	    >Calculate end of file marker and first free byte.
C	>If file has been extended and End-of-file is short of 
C	 allocated blocks (+10).
C	    >Truncate file.
C		>If not successful:
C		    >Return error status.
C  >Deaccess file - writing file attributes.
C 	>If not successful:
C	    >Return with error status.
C  >Deassign channel.
C	>If not successful:
C	    >Return error status.
C  >Return error status.

	IF (MAP$ACCESS_TYPE .EQ. MAP$C_RW) THEN

		IF (FILELENGTH .NE. MAP$FILE_LEN) THEN
			MAP$EOF_BLK_NO = FILELENGTH/512
			MAP$FREE_BYTE  = FILELENGTH - MAP$EOF_BLK_NO*512
			IF (MAP$FREE_BYTE .GT. 0) MAP$EOF_BLK_NO =
	1				MAP$EOF_BLK_NO + 1
		ENDIF

		IF ( MAP$ALLOC_BLKS .GT. MAP$EOF_BLK_NO + 10) THEN
			FIB$L_ACCTL = FIB$M_WRITE
			FIB$W_NMCTL = FIB$M_FINDFID
			FIB$W_EXCTL = FIB$M_TRUNC
			FIB$L_EXVBN = MAP$EOF_BLK_NO + 1
			FIB$L_EXSZ  = 0
C			ISTATUS = SYS$QIOW(
C	1			,%VAL(MAP$CHANNEL)
C	1			,IO$_MODIFY
C	1			,IOSB
C	1			,,,MAP$FIB_DESCR
C	1			,,,,,)
C			IF (.NOT. ISTATUS .OR. .NOT. IOSB_W(1)) THEN
C				VMP$CLOSE = IOSB_W(1)
C				RETURN
C			ENDIF
C	TYPE *,'TRUNCATION should have occurred here.'
		ENDIF
	ENDIF

	MAP$W_REC_EOF  = MAP$EOF_BLK_NO
	MAP$W_REC_FREE = MAP$FREE_BYTE
	FIB$W_EXCTL = 0
	FIB$L_EXSZ  = 0
	FIB$L_EXVBN = 0
	FIB$W_NMCTL = FIB$M_FINDFID
	ISTATUS = SYS$QIOW(
	1	,%VAL(MAP$CHANNEL)
	1	,IO$_DEACCESS
	1	,IOSB
	1	,,,MAP$FIB_DESCR
	1	,,,,ATR$L_LIST
	1	,)
	IF (.NOT. ISTATUS .OR. .NOT. IOSB_W(1)) THEN
		CALL REPORT('VMP$CLOSE .. failure to close file '''
	1	  //NCODE(IOSB_W(1))//'''X')
		VMP$CLOSE = VMP$_CLSFIL
		RETURN
	ENDIF
	
	IF (MAP$ACCESS_TYPE .EQ. MAP$C_RW 
	1	.AND. MAP$ALLOC_BLKS .GT. MAP$EOF_BLK_NO + 8) THEN
		ISTATUS = VMP$TRUNCATE()
		IF (.NOT. ISTATUS) THEN
			CALL REPORT('VMP$CLOSE .. failure to truncate file ')
			VMP$CLOSE = VMP$TRNFAI
			RETURN
		ENDIF
	ENDIF

	ISTATUS = SYS$DASSGN (%VAL(MAP$CHANNEL))
	IF (.NOT. ISTATUS) THEN
		CALL REPORT('VMP$CLOSE .. failure to deassign channel '''
	1	  //NCODE(ISTATUS)//'''X')
		VMP$CLOSE = VMP$_CLSFIL
		RETURN
	ENDIF


	VMP$CLOSE = %LOC(SS$_NORMAL)
	RETURN
	END

C********** VMP$EXTEND - extend file **********************

	INTEGER*4	FUNCTION VMP$EXTEND()

	INCLUDE		'VMPLIBDEF.FOR/NOLIST'

	EXTERNAL	SS$_NORMAL,IO$_MODIFY
	INTEGER*4	SS$_NORMAL,IO$_MODIFY
	INTEGER*4	SYS$QIOW
	INTEGER*4	ISTATUS			    ! local.



C********** Explanation of local symbols ***********
C
C	ISTATUS		(local) Program status longword.



C********** Extend file via ACP calls ***********
C
C  >If file-access is readonly:
C	>Return error status that file is readonly.
C  >Request file extension - best effort at contiguous allocation.
C  >Request extension at end of file (0).
C  >Perform QIOW to current channel.
C  >If extension request was not successful:
C	>Return failure status.
C  >Update file descriptor block:number of actual blocks allocated,no. extends
C  >Update FIB:allocated blocks,eof block no.,and free byte.
C  >Update file header with end-of-file information - QIOW.
C  >If not successful:
C	>Return error status
C  >If successful:
C	>Return true status.

	IF (MAP$ACCESS_TYPE .NE. MAP$C_RW) THEN
		VMP$EXTEND = VMP$E_NOWRITE
		RETURN
	ENDIF

	FIB$W_NMCTL = FIB$M_FINDFID	
	FIB$W_EXCTL = FIB$M_EXTEND .OR. FIB$M_ALCONB  
	FIB$L_EXVBN = MAP$ALLOC_BLKS + 1
	FIB$L_EXSZ = MAP$WINDOW_BLKS	

	ISTATUS = SYS$QIOW(
	1		,%VAL(MAP$CHANNEL)
	1		,IO$_MODIFY		
	1		,IOSB
	1		,,
	1		,MAP$FIB_DESCR
	1		,,,
	1		,
	1		,)
	IF (.NOT. ISTATUS .OR. .NOT. IOSB_W(1))
	1   THEN
		CALL REPORT('VMP$EXTEND .. failure to extend file '''
	1	  //NCODE(IOSB_W(1))//'''X')
		VMP$EXTEND = VMP$_EXTFAI
		RETURN
	ENDIF

	MAP$ALLOC_BLKS = MAP$ALLOC_BLKS + FIB$L_EXSZ
	MAP$NO_EXTENDS = MAP$NO_EXTENDS + 1
	MAP$EXT_FILELEN = MAP$ALLOC_BLKS*512 ! EXT_FILELEN + FIB$L_EXSZ*512
	MAP$W_REC_ALLOC = MAP$ALLOC_BLKS
	MAP$W_REC_EOF   = MAP$ALLOC_BLKS + 1
	MAP$W_REC_FREE  = MAP$FREE_BYTE

	FIB$W_EXCTL = 0
	FIB$L_EXVBN = 0
	FIB$L_EXSZ  = 0

	ISTATUS = SYS$QIOW(
	1		,%VAL(MAP$CHANNEL)
	1		,IO$_MODIFY			
	1		,IOSB
	1		,,
	1		,MAP$FIB_DESCR
	1		,,,
	1		,ATR$L_LIST
	1		,)

	IF (ISTATUS .NE. %LOC(SS$_NORMAL) .OR. IOSB_W(1) .NE.%LOC(SS$_NORMAL))
	1	THEN
		CALL REPORT('VMP$EXTEND .. failure to update file header '''
	1	  //NCODE(IOSB_W(1))//'''X')	
		VMP$EXTEND = VMP$E_ERRHDRUPD
	ELSE
		VMP$EXTEND = %LOC(SS$_NORMAL)
	ENDIF

	RETURN
	END

C********** VMP$UPDATE *********************************

	INTEGER*4	FUNCTION VMP$UPDATE()

	INCLUDE		'VMPLIBDEF.FOR/NOLIST'

	PARAMETER	EFN = 3
	EXTERNAL	SS$_NORMAL,SS$_NOTMODIFIED
	INTEGER*4	SS$_NORMAL,SS$_NOTMODIFIED,SYS$UPDSEC
	INTEGER*4	ISTATUS			! local



C********** Explanations of locals ***********
C
C	ISTATUS		= local - program status longword.



C********** Update section file - writing modified pages to disk **********
C
C  >Request update of program section - if any pages have been modified
C	one or more QIOWs are issued to update section file on disk.
C  >If normal completion:
C	>Wait for event flag that signals that all QIOWs are complete.
C	>Return any error status from update section.
C  >If warning that no pages were modified:
C	>Return immediately.
C  >Return error status

	VMP$UPDATE = %LOC(SS$_NORMAL)

	ISTATUS = SYS$UPDSEC(MAP$MAP_ADDR,,%VAL(3),%VAL(1)	
	1	,%VAL(EFN),IOSB,,)
	IF (ISTATUS .EQ. %LOC(SS$_NORMAL)) THEN
		CALL SYS$WAITFR(%VAL(EFN))
		IF (.NOT. IOSB_W(1)) VMP$UPDATE = IOSB(1)
	ELSE
		IF (ISTATUS .NE. %LOC(SS$_NOTMODIFIED))
	1		VMP$UPDATE = ISTATUS
	ENDIF

	RETURN
	END

C********* VMP$TRUNCATE - truncate file ***********************

	INTEGER*4	FUNCTION VMP$TRUNCATE

	INCLUDE		'VMPLIBDEF.FOR/NOLIST'

	INTEGER*4	ISTATUS				!local

	EXTERNAL	SS$_NORMAL,IO$_ACCESS,IO$_DEACCESS,IO$_MODIFY
	EXTERNAL	IO$M_ACCESS
	INTEGER*4	SS$_NORMAL,IO$_ACCESS,IO$_DEACCESS,IO$_MODIFY
	INTEGER*4	IO$M_ACCESS
	INTEGER*4	SYS$QIOW

	INTEGER*4	SYS$ASSIGN


C********** Explanation of locals *********
C
C	ISTATUS		(local).  Program module error status.



C********** Truncate file **********
C
C NOTE:  This is temporary - currently there is a bug in the ACP
C	 software (version 1.6) where a truncation request on a
C	 file that has been accessed and then mapped to causes
C	 a CONFLICT error, and truncation after the file has been
C	 deaccessed, causes the VAX to crash.
C
C
C  >Access file.
C    >If not successful, return error status.
C  >Truncate file.
C    >If not successful, return error status.
C  >Deaccess file.
C    >If not successful, return error status.

	FIB$L_ACCTL = FIB$M_WRITE
	FIB$W_EXCTL = 0
	FIB$L_EXSZ  = 0
	FIB$L_EXVBN = 0
	FIB$W_NMCTL = FIB$M_FINDFID

	ISTATUS = SYS$QIOW(,%VAL(MAP$CHANNEL)
	1		,%VAL(%LOC(IO$_ACCESS).OR.%LOC(IO$M_ACCESS))
	1		,IOSB,,,MAP$FIB_DESCR
	1		,,,,ATR$L_LIST,)
	IF (.NOT. ISTATUS .OR. .NOT. IOSB_W(1)) THEN
		CALL REPORT('VMP$TRUNCATE .. failed to access file '''
	1		//NCODE(IOSB_W(1))//'''X')
		VMP$TRUNCATE = VMP$_TRNFAI
		RETURN
	ENDIF

	FIB$W_NMCTL = FIB$M_FINDFID
	FIB$W_EXCTL = FIB$M_TRUNC
	FIB$L_EXSZ  = 0
	FIB$L_EXVBN = MAP$W_REC_EOF + 1
	ISTATUS = SYS$QIOW(,%VAL(MAP$CHANNEL),%VAL(%LOC(IO$_MODIFY))
	1		,IOSB,,,MAP$FIB_DESCR
	1		,,,,,)
	IF (.NOT. ISTATUS .OR. .NOT. IOSB_W(1)) THEN
		CALL REPORT('VMP$TRUNCATE .. failed to truncate file '''
	1		//NCODE(IOSB_W(1))//'''X')
		VMP$TRUNCATE = VMP$_TRNFAI
		RETURN
	ENDIF

	FIB$W_EXCTL = 0
	FIB$L_EXVBN = 0
	FIB$L_EXSZ = 0
	ISTATUS = SYS$QIOW(,%VAL(MAP$CHANNEL),%VAL(%LOC(IO$_DEACCESS))
	1		,IOSB,,,MAP$FIB_DESCR
	1		,,,,,)
	IF (.NOT. ISTATUS .OR. .NOT. IOSB_W(1)) THEN
		CALL REPORT('VMP$TRUNCATE .. failed to deaccess file '''
	1		//NCODE(IOSB_W(1))//'''X')
		VMP$TRUNCATE = VMP$_TRNFAI
		RETURN
	ENDIF

	VMP$TRUNCATE = %LOC(SS$_NORMAL)
	RETURN
	END	

C*********** VMP$CR8_FILDSCR *******************************************
C
C	VMP$GET_FILDSCR expands the users virtual address space by
C	a single block and then calls a routine to initialize that
C	space so that it can be used as a file descriptor block. The
C	routine returns the logical id number by which the file
C	should be referenced by during calls to VMP$MAP_RECORD, and
C	VMP$UNMAP_FILE.


	INTEGER*4	FUNCTION VMP$CR8_FILDSCR(FILEID)

	INCLUDE		'VMPLIBDEF.FOR/NOLIST'

	EXTERNAL	SS$_NORMAL
	INTEGER*4	SYS$EXPREG,SS$_NORMAL

	INTEGER*4	FILEID				! FILE ID NUMBER
	INTEGER*4	ISTATUS 	! local

	INTEGER*4	VALUE_OF,STORE_VALUE



C********** Explanation of parameters and locals **********
C
C	FILEID		= integer number returned by CR8_FILDSCR
C			  to be used as an identifier of the file.
C	ISTATUS 	= local - program status longword.


			  
C********* Get virtual memory and initialize it as a file descriptor block ***
C
C  >Request one block of virtual address space.
C  >If request fails:
C	>Return error status.
C  >Start at beginning of linked list.
C  >Loop:
C	>If end of list:
C		>Store address of file_descriptor block
C		>Calculate fileid - id of last block plus 1
C	>If not end of list:
C		>Save fileid of this block
C		>Point at next descriptor block.
C		>Go back to start of loop.
C  >Initialize descriptor block.
C  >Return true status.

	ISTATUS = SYS$EXPREG 	(%VAL(FILE_DESCR_BLK)
	1			,RETADR
	1			,%VAL(3)		!ACCESS MODE - USER
	1			,%VAL(0)		!P0 region
	1			)

	IF ( .NOT. ISTATUS) THEN
		VMP$CR8_FILDSCR = ISTATUS
		RETURN
	ENDIF

	ITEMP = 0
	PREV_PTR = 0
	CURR_PTR = %LOC(MAP_PTR)

  20	IF ( VALUE_OF(%VAL(CURR_PTR)) .EQ. 0)
	1   THEN
		CALL STORE_VALUE(%VAL(CURR_PTR),RETADR(1))
		FILEID = ITEMP+ 1
	   ELSE
		PREV_PTR = CURR_PTR
		ITEMP = VALUE_OF(%VAL(VALUE_OF(%VAL(CURR_PTR))+4))
		CURR_PTR = VALUE_OF(%VAL(CURR_PTR))
		GOTO 20
	ENDIF

	VMP$CR8_FILDSCR = %LOC(SS$_NORMAL)

	RETURN
	END

C********** VMP$INIT_DESCR **********************************************
C
C	VMP$INIT_DESCR  initializes the fields of the file descriptor
C	block.

	INTEGER*4 FUNCTION VMP$INIT_DESCR(TYPE,ACCESS,WINDOWSIZE)

	INCLUDE		'VMPLIBDEF.FOR/NOLIST'

	INTEGER*4	WINDOWSIZE
	CHARACTER*(*)	TYPE
	CHARACTER*(*)	ACCESS

	INTEGER*4	ISTATUS			! local

	EXTERNAL	SS$_NORMAL
	INTEGER*4	SS$_NORMAL


C*********** Explanation of parameters and locals **********
C
C	TYPE		= 'NEW' or 'OLD' - character string for
C			  file type.
C	ACCESS		= 'R/W' read/write or 'R/O' readonly -
C			   character string - access mode.
C	WINDOWSIZE	= size of window in bytes - integer*4.
C			  window will actually be the number of
C			  blocks needed to accomodate this.
C			  (It must be greater than 0).
C	ISTATUS		= local - program status longword.



C*********** Initialize file descriptor **********
C
C  >Zero all fields.
C  >Store filelength.
C  >Store file type - old or new
C  >Store access type - read/write or readonly
C  >Store fileid
C  >Calculate number of blocks in window.
C  >Initialize descriptor for fib.
C  >Initialize attribute list.

	IF ( (TYPE(1:3) .NE. 'NEW' .AND. TYPE (1:3) .NE. 'OLD')
	1  .OR.  (ACCESS(1:3) .NE. 'R/O' .AND. ACCESS(1:3) .NE. 'R/W')
	1  .OR.  (TYPE(1:3) .EQ. 'NEW' .AND. ACCESS(1:3) .EQ. 'R/O')
	1  .OR.  WINDOWSIZE .LE. 0) THEN
		VMP$INIT_DESCR = VMP$E_INVPAR
		RETURN
	ENDIF

	DO 30 ITEMP = 0, FILE_DESCR_LLEN ! ?? CAN THIS BE DONE BY EXPREG
   30	FILE_DESCR_L(ITEMP) = 0

	MAP$NO_EXTENDS = 0
	MAP$TYPE = MAP$C_OLD
	IF (TYPE(1:3) .EQ. 'NEW') MAP$TYPE = MAP$C_NEW
	MAP$ACCESS_TYPE = MAP$C_RW	!
	IF (ACCESS(1:3) .EQ. 'R/O') MAP$ACCESS_TYPE = MAP$C_RO
					! from argument list

	MAP$WINDOW_BLKS = WINDOWSIZE/512
	IF (MAP$WINDOW_BLKS*512 .NE. WINDOWSIZE)
	1	MAP$WINDOW_BLKS = MAP$WINDOW_BLKS + 1
	MAP$WINDOW_SIZE = MAP$WINDOW_BLKS * 512

	MAP$FIB_DESCR(1) = FIB$K_LEN
	MAP$FIB_DESCR(2) = %LOC(MAP$FIB)

	ATR$W_LIST(1)	= ATR$S_RECATTR
	ATR$W_LIST(2) 	= ATR$C_RECATTR
	ATR$L_LIST(2) 	= %LOC(MAP$B_RECATTR)

	ATR$L_LIST(3)	= ATR$K_EOLIST

	VMP$INIT_DESCR = %LOC(SS$_NORMAL)

	RETURN
	END

C********** VMP$COPY **************************************
C
C	translates to a MOVC5 !!!!

	SUBROUTINE 	VMP$COPY(INSTRING,OUTSTRING)

	CHARACTER*(*)	INSTRING,OUTSTRING

	OUTSTRING = INSTRING
	RETURN
	END

C********** VMP$GET_DESCR ***********************************

	INTEGER*4	FUNCTION VMP$GET_FILDSCR(ID)

	INCLUDE		'VMPLIBDEF.FOR/NOLIST'

	INTEGER*4	ID		! parameter
	INTEGER*4	ISTATUS		! local

	EXTERNAL	SS$_NORMAL
	INTEGER*4	SS$_NORMAL
	INTEGER*4	VALUE_OF



C********** Explanation of parameters and local symbols **********
C
C	ID		(parameter). Internal file id - identifies
C			which file descriptor to retrieve.
C
C	ISTATUS		(local). Program status longword.




C********** Get file descriptor block by id number **********
C
C  >Return if the current file descriptor block is the one requested.
C  >Set up pointers to beginning of linked list
C  >Loop through linked list of descriptor blocks:
C  >If requested id matches:
C	>Copy descriptor block into temp area and return true status.
C  >If end of linked list:
C	>Return false status.
C	
C
	VMP$GET_FILDSCR = %LOC(SS$_NORMAL)
	IF (VALUE_OF(%VAL(VALUE_OF(%VAL(CURR_PTR))+4)) .EQ. ID) RETURN

	PREV_PTR = 0
	CURR_PTR = %LOC(MAP_PTR)
20	IF (VALUE_OF(%VAL(CURR_PTR)) .EQ. 0) THEN
		VMP$GET_FILDSCR =VMP$_NOTMAP
		RETURN
	ELSE
		IF (VALUE_OF(%VAL(VALUE_OF(%VAL(CURR_PTR))+4)) .EQ. ID)
	1			GOTO 50
		PREV_PTR = CURR_PTR
		CURR_PTR =VALUE_OF(%VAL(CURR_PTR))
		GOTO 20
	ENDIF

50	DESCR(1) = FILE_DESCR_LEN
	DESCR(2) = VALUE_OF(%VAL(CURR_PTR))
	CALL VMP$COPY(DESCR,FILE_DESCR_C)
	VMP$GET_FILDSCR = %LOC(SS$_NORMAL)
	RETURN
	END

C*********** VALUE_OF - move contents of parameter into r0 **********
C
C	VALUE_OF is usually used to help move through a linked
C	list in fortran.  It is usually invoked:
C		VAL = VALUE_OF(%VAL(LOCATION))

	INTEGER*4 FUNCTION	VALUE_OF(ARRAY)
	INTEGER*4 ARRAY(1)

	VALUE_OF = ARRAY(1)
	RETURN
	END

C*********** STORE_VALUE - move contents of one location to another *********
C
C	STORE_VALUE is used to store indirect and is invoked:
C		CALL STORE_VALUE(%VAL(CURR_PTR),NUMBER)

	INTEGER*4 FUNCTION	STORE_VALUE(LOCATION,VALUE)
	INTEGER*4	LOCATION,VALUE

	LOCATION = VALUE
	RETURN
	END

C********** REPORT - Report severe error **********

	SUBROUTINE REPORT(MESSAGE)
	CHARACTER*(*) MESSAGE					! Parameter

	INCLUDE		'VMPLIBDEF.FOR/NOLIST'



C********** Explanation of parameters and locals **********
C
C  MESSAGE =		(*)-character message describing error.




C********** Reporting of severe error **********
C
C  >Type severe error report on terminal:
C	>Ring bell.
C	>Type heading.
C	>Type message describing error.


	WRITE (SYS$OUTPUT,104) 7
104	FORMAT (X,A)

	WRITE (SYS$OUTPUT,105) MESSAGE
105	FORMAT (/' --- VMP$ program severe error ---'/
	2        '          Message:  ',A/)
	RETURN
	END



C********** NCODE - Code error status into external format **********

	CHARACTER*(*)	FUNCTION NCODE(ERROR_NUMBER)
	INTEGER*4	ERROR_NUMBER		! Parameter
	

C********** Explanation of parameters **********
C
C  ERROR_NUMBER		= integer*4 number containing error code
C


	ENCODE(8,10,NCODE)ERROR_NUMBER
10	FORMAT(Z8)

	DO 20 I = 1,8
	IF (NCODE(I:I) .EQ. ' ') NCODE(I:I) = '0'
20	CONTINUE

	RETURN
	END

C********** LENGTH - returns length of string **********
C
C	LENGTH returns actual length of string - that is length
C	not including trailing blanks and/or nulls.

	INTEGER*4 FUNCTION LENGTH(STR)
	CHARACTER*(*) STR
	CHARACTER*1	NULL
	BYTE		NULL_B(1)
	EQUIVALENCE	(NULL,NULL_B(1))
	DATA		NULL_B /0/
	LENT = LEN(STR)
	DO 10 I = LENT,1,-1
		LENGTH = I
		IF(STR(I:I) .NE. ' ' .AND. STR(I:I) .NE. NULL) GO TO 20
10	CONTINUE
	LENGTH = 0
20	RETURN
	END
